/*
  interpolation.cc
  Author: Luca Rottoli, 2015
  modified by Marco Bonvini, 2017
*/

#include "CoefficientFunctionsOffShellInterpolator.hh"

using namespace std;

namespace HELL {

  dcomplex Gamma(dcomplex z){
    gsl_sf_result r, arg;
    gsl_sf_lngamma_complex_e(real(z),imag(z),&r,&arg);  // causes problems in (x,0), x<0
    return polar(exp(r.val),arg.val);
  }
  
  
  double Interpolator::xi_sample(int i) {
    double interval = 1./(fNPoints-1);
    return exp(log(ximin/2) + log(4*ximax/ximin)*interval*i);
    return pow(0.+ interval*i,3);
  }


#define nn 7
  const string names[] = {"2", "L", "2ax", "Lax", "2cc", "Lcc", "3cc"};

  /*
	Constructor
  */
  Interpolator::Interpolator(double xim, unsigned int npoints)
  {
    fNPoints = npoints;
    fxim = xim;
    f_slope = 0.;
    f_reM0 = 1./2;
    ximin = 2*fxim/1e5;
    ximax = ( fxim>1./4. ? 4*fxim*1e5 : 1e5 );
    //
    double * xia = new double [fNPoints];
    //double * xib = new double [fNPoints];
    map<string,double*> Capoints;//, Cbpoints;
    //  
    for(int k=0; k<nn; k++) {
      fSplineCa[names[k]] = gsl_spline_alloc (gsl_interp_cspline, fNPoints);
      //fSplineCb[names[k]] = gsl_spline_alloc (gsl_interp_cspline, fNPoints);
      Capoints [names[k]] = new double [fNPoints];
      //Cbpoints [names[k]] = new double [fNPoints];
    }
    
    cout << "Initialization of off-shell cross sections...";
    cout.flush();
    
    for(int i = 0; i < fNPoints; i++) {
      xia[i] = xi_sample(i);
      /*
      for(int k=0; k<4; k++) {
	Capoints[names[k]][i] = InverseMellinC(names[k],xia[i]);
      }
      for(int k=4; k<nn; k++) {
	Capoints[names[k]][i] = log(-InverseMellinC(names[k],xia[i]));
      }
      */
      for(int k=0; k<nn; k++) {
	Capoints[names[k]][i] = log(-InverseMellinC(names[k],xia[i]));
      }
      //cout << setw(12) << xia[i] << setw(15) << Capoints["Lcc"][i] << endl;
    }
    // hack for removing the inf from the first point
    /*
    for(int k=4; k<nn; k++) {
      Capoints[names[k]][0] = log(-Csmallxi(names[k],xi_sample(1)/2));
    }
    */
    /*
    for(int i = 0; i < fNPoints; i++) {
      xib[i] = 1./xia[fNPoints-1-i];
      for(int k=0; k<nn; k++) {
	Cbpoints[names[k]][i] = log(-InverseMellinC(names[k],xib[i]));
      }
      //cout << setw(12) << xib[i] << setw(15) << Cbpoints["L"][i] << endl;
    }
    // hack for.. ?
    for(int k=0; k<nn; k++) {
      Cbpoints[names[k]][fNPoints-1] = 1e100;
    }
    */
    for(int k=0; k<nn; k++) {
      gsl_spline_init(fSplineCa[names[k]], xia, Capoints[names[k]], fNPoints);
      //gsl_spline_init(fSplineCb[names[k]], xib, Cbpoints[names[k]], fNPoints);
    }
    
    cout << " completed!" << endl;
    
    for(int k=0; k<nn; k++) {
      delete[] Capoints[names[k]];
      //delete[] Cbpoints[names[k]];
    }
    delete [] xia;
    //delete [] xib;
  }

/*
	Destructor
 */
  Interpolator::~Interpolator()
  {
    for(int k=0; k<nn; k++) {
      gsl_spline_free(fSplineCa[names[k]]);
      //delete fSplineCa[names[k]];
      //delete fSplineCb[names[k]];
    }
  }
  
/*
  Copy Constructor
*/
  Interpolator::Interpolator(Interpolator& other):
    fNPoints(other.fNPoints)
  {
    fNPoints = other.fNPoints;
    fxim     = other.fxim;
    f_reM0   = other.f_reM0;
    f_slope  = other.f_slope;
    for(int k=0; k<nn; k++) {
      fSplineCa[names[k]] = new gsl_spline;
      //fSplineCb[names[k]] = new gsl_spline;
      fSplineCa[names[k]] = other.fSplineCa[names[k]];
      //fSplineCb[names[k]] = other.fSplineCb[names[k]];
    }
  }

/*
  C2(xi) returns the value of C2 at xi
 */
  double Interpolator::C(string type, double xi) {
    if(xi==1./0) return 0;
    if(xi<ximin) return Csmallxi(type, xi);
    if(xi>ximax) return Clargexi(type, xi);
    double result = 0;
    gsl_interp_accel *acc = gsl_interp_accel_alloc ();
    /*
    if(xi <= 1 ) {
      if(type=="2" || type=="L" || type=="2ax" || type=="Lax") result =      gsl_spline_eval(fSplineCa[type], xi , acc);
      if(type=="2cc" || type=="Lcc" || type=="3cc")            result = -exp(gsl_spline_eval(fSplineCa[type], xi , acc));
    }
    if(xi >  1 ) result = -exp(gsl_spline_eval(fSplineCb[type], xi , acc));
    */
    result = -exp(gsl_spline_eval(fSplineCa[type], xi , acc));
    gsl_interp_accel_free (acc);
    //delete acc;
    return result;
  }


  // all these results are in study_catani.nb
  double Interpolator::Csmallxi(string type, double xi) {
    gsl_complex gslcomp = gsl_complex_rect (sqrt(4*fxim), 0);
    double ArcCsch = GSL_REAL (gsl_complex_arccsch(gslcomp));
    double res = 0;
    if(type == "2"  ) res = -(4.*(sqrt(1 + 4*fxim) + 4*fxim*ArcCsch))/(5.*4*fxim*pow(1 + 4*fxim,1.5));
    if(type == "2ax") res = -1/(1+4*fxim) - 4*fxim*ArcCsch/pow(1+4*fxim,1.5);
    if(type == "L"  ) res = (3*sqrt(1. + 4*fxim)*(2 + 4*fxim) - (8. + 4*fxim*(8 + 3.*4*fxim))*ArcCsch)/(10.*pow(1 + 4*fxim,2.5));
    if(type == "Lax") res = -(4*fxim/(1+4*fxim))*(3*sqrt(1. + 4*fxim)*(1 + 3*fxim) + 2*fxim*(7+22*fxim)*ArcCsch)/(10*fxim*pow(1+4*fxim,1.5));
    if(type == "2cc") res = -1./xi/2.;
    if(type == "Lcc") res = -fxim/(1+fxim)/xi/2.;
    if(type == "3cc") res = -1./(1+fxim)/xi/2.;
    return res;
  }
  double Interpolator::Clargexi(string type, double xi) {
    //double eps = 1e-5;
    double res = 0;
    if(type == "2"  ) {
      //const complex<double> prime  = (hyp_2F1 (1,0.+eps,3./2,-1./fxim/4.)-hyp_2F1 (1,0.-eps,3./2,-1./fxim/4.))/(2.*eps);
      //double twoFonePrime = prime.real();
      //res = -1./pow(xi,2)*(-5./6+log(xi/fxim)+(1.-2*fxim)*twoFonePrime);
      double sq = sqrt(1+4*fxim);
      res = 1./pow(xi,2)/6*(-7+24*fxim-6*log(xi/fxim)+6*(2*fxim-1)*sq*log((sq-1)/(sq+1)));
    }
    if(type == "2ax") {
      //const complex<double> prime  = (hyp_2F1 (1,0.+eps,-1./2,-1./fxim/4.)-hyp_2F1 (1,0.-eps,-1./2,-1./fxim/4.))/(2.*eps);
      //double twoFonePrime = prime.real();
      //res = 6*fxim/pow(xi,2)*(1+8*fxim-log(xi/fxim)-pow(1+4*fxim,2)*twoFonePrime);
      double sq = sqrt(1+4*fxim);
      res = -6*fxim/pow(xi,2)*(1+log(xi/fxim)+sq*log((1+2*fxim-sq)/(2*fxim)));
    }
    if(type == "L"  ) {
      //const complex<double> prime  = (hyp_2F1 (1,0.+eps,3./2,-1./fxim/4.)-hyp_2F1 (1,0.-eps,3./2,-1./fxim/4.))/(2.*eps);
      //double twoFonePrime = prime.real();
      //res = -1./pow(xi,2)*(1./(1.+4*fxim))*(1.-fxim*2.-3./4.*pow(4*fxim,2)*twoFonePrime);
      double sq = sqrt(1+4*fxim);
      res = -1./pow(xi,2)*(1-6*fxim-12*fxim*fxim/sq*log((sq-1)/(sq+1)));
    }
    if(type == "Lax") {
      gsl_complex gslcomp = gsl_complex_rect (sqrt(4*fxim), 0);
      double ArcCsch = GSL_REAL (gsl_complex_arccsch(gslcomp));
      //const complex<double> primeA  = (hyp_2F1 (1,1.+eps,3./2,-1./fxim/4.)-hyp_2F1 (1,1.-eps,3./2,-1./fxim/4.))/(2.*eps);
      //double twoFonePrimeA = primeA.real();
      //const complex<double> primeB  = (hyp_2F1 (2,1.+eps,5./2,-1./fxim/4.)-hyp_2F1 (2,1.-eps,5./2,-1./fxim/4.))/(2.*eps);
      //double twoFonePrimeB = primeB.real();
      //res = 6*fxim/pow(xi,2)*sqrt(1+4*fxim)*ArcCsch - (1+4*fxim)/4/xi/xi/fxim*(6*fxim*(log(xi/fxim)+twoFonePrimeA) + twoFonePrimeB);
      double sq = sqrt(1+4*fxim);
      res = -3*sq/2./pow(xi,2)*(4*fxim/(1+4*fxim))*(-4*fxim*ArcCsch + (1+4*fxim)*log((sq-1)/(sq+1)) + sq*(2+log(xi/fxim)));
    }
    if(type == "2cc") {
      res = 1/pow(xi,2)/6*(-7+3*fxim*(1+fxim)-3*(2+3*fxim)*log(xi)+3*fxim*fxim*fxim*log(fxim)-3*(fxim-2)*pow(1+fxim,2)*log(1+fxim));
    }
    if(type == "Lcc") {
      res = 1/pow(xi,2)/2*(-2+3*fxim*fxim+3*fxim*(-log(xi)+fxim*fxim*log(fxim/(1+fxim))+log(1+fxim)));
    }
    if(type == "3cc") {
      //const complex<double> primeA  = (hyp_2F1(1,4.+eps,3.,1./(1+fxim))-hyp_2F1(1,4.-eps,3.,1./(1+fxim)))/(2.*eps);
      //double twoFonePrimeA = primeA.real();
      //const complex<double> primeB  = (hyp_2F1(1,4.,3.+eps,1./(1+fxim))-hyp_2F1(1,4.,3.-eps,1./(1+fxim)))/(2.*eps);
      //double twoFonePrimeB = primeB.real();
      //res = 2*fxim/(1+fxim)/pow(xi,3)*(-(1+fxim)*log(xi/fxim)+3*fxim*fxim*(2*twoFonePrimeA+twoFonePrimeB));
      res = -fxim/pow(xi,3)/3*(7+3*fxim*(5+2*fxim)+6*log(xi/fxim)+6*pow(1+fxim,3)*log(fxim/(1+fxim)));
    }
    return res;
  }
  /*
  dcomplex Interpolator::singMellin(string type, dcomplex M) {
    dcomplex result = 0.;
    //if(type == "2cc") result = Gamma(M)/2.;
    //if(type == "Lcc") result = Gamma(M)/2. * fxim/(1+fxim);
    //if(type == "3cc") result = Gamma(M)/2. / (1+fxim);
    return result;
  }
  double Interpolator::sing(string type, double xi) {
    double result = 0.;
    //if(type == "2cc") result = Gamma(M)/2.;
    //if(type == "Lcc") result = Gamma(M)/2. * fxim/(1+fxim);
    //if(type == "3cc") result = Gamma(M)/2. / (1+fxim);
    return result;
  }
  */
  dcomplex Interpolator::mellinC(string type, dcomplex M) {
    dcomplex result;
    complex<double> MM;
    MM.real(real(M));
    MM.imag(imag(M));
    if(type == "2"  ) {
      const complex<double> F3  = hyp_2F1 (1.-MM, 1, 3./2, -1./fxim/4.);
      dcomplex twoFone3 (F3 .real(),F3 .imag());
      result =  3./2*pow(fxim,M)*pow(Gamma(1.-M),3)*Gamma(1.+M)/((3.-2.*M)*(1.+2.*M)*Gamma(2.-2.*M))*(1.+M-(1.+M-(2.+3.*M-3.*M*M)/fxim/2.)*twoFone3);
    }
    if(type == "2ax") {
      const complex<double> Fm1 = hyp_2F1 (1.-MM, 1,-1./2, -1./fxim/4.);
      dcomplex twoFonem1(Fm1.real(),Fm1.imag());
      result = -3./2*pow(fxim,M)*pow(Gamma(1.-M),3)*Gamma(1.+M)/((3.-2.*M)*(1.+2.*M)*Gamma(2.-2.*M))*2.*(1.+2.*M)/(1.-2.*M)*(8*fxim*(M-2*fxim-2.)+pow(4*fxim+1,2)*twoFonem1);
    }
    if(type == "L"  ) {
      const complex<double> F3  = hyp_2F1 (1.-MM, 1, 3./2, -1./fxim/4.);
      dcomplex twoFone3 (F3 .real(),F3 .imag());
      result =  3./2*pow(fxim,M)*pow(Gamma(1.-M),3)*Gamma(1.+M)/((3.-2.*M)*(1.+2.*M)*Gamma(2.-2.*M))*(4*fxim/(1+4*fxim))*(3.+(1.-M)/fxim/2.-(3.+(1.-M)/fxim*(1.-M/fxim/4.))*twoFone3);
    }
    if(type == "Lax") {
      const complex<double> F3a = hyp_2F1 (2.-MM, 1, 3./2, -1./fxim/4.);
      const complex<double> F5a = hyp_2F1 (2.-MM, 2, 5./2, -1./fxim/4.);
      dcomplex twoFone3a(F3a.real(),F3a.imag());
      dcomplex twoFone5a(F5a.real(),F5a.imag());
      result = -3./2*pow(fxim,M)*pow(Gamma(1.-M),3)*Gamma(1.+M)/((3.-2.*M)*(1.+2.*M)*Gamma(2.-2.*M))*(1.+2.*M)/3./fxim*(6.*fxim*(2.*M-3.)*twoFone3a+(3.*M-4.)*twoFone5a);
    }
    if(type == "2cc") {
      const complex<double> F = hyp_2F1 (2.*MM-1., 1, MM+1., 1./(1.+fxim));
      dcomplex twoFone(F.real(),F.imag());
      result = -3./4*pow(fxim,M)*pow(Gamma(1.-M),3)*Gamma(1.+M)/(M*(3.-2.*M)*(1.+2.*M)*(1.-2.*M)*Gamma(2.-2.*M)*fxim)*(M*(M*M*(fxim*fxim-2*fxim-3)-M*(fxim*fxim-4*fxim-3)+fxim+2.)+(1.-M)*(1+fxim)*(M*M*(fxim+3)-3.*M-2.)*twoFone);
    }
    if(type == "Lcc") {
      const complex<double> F = hyp_2F1 (2.*MM-1., 1, MM+1., 1./(1.+fxim));
      dcomplex twoFone(F.real(),F.imag());
      result =  3./4*pow(fxim,M)*pow(Gamma(1.-M),3)*Gamma(1.+M)/(M*(3.-2.*M)*(1.+2.*M)*(1.-2.*M)*Gamma(2.-2.*M)*fxim)*(2.*M*M*M*(3*fxim+1)-M*M*(fxim*fxim+11*fxim+2)-M*fxim*(fxim-2)+2.*fxim*fxim+(M-1.)*(2.*M*M*(fxim+1)+M*(fxim*fxim-5*fxim-2)+2.*fxim*fxim)*twoFone);
    }
    if(type == "3cc") {
      const complex<double> F = hyp_2F1 (2.*MM, 1, MM+1., 1./(1.+fxim));
      dcomplex twoFone(F.real(),F.imag());
      result =  3./2*pow(fxim,M)*(1.-M)*pow(Gamma(1.-M),3)*Gamma(1.+M)/(M*(3.-2.*M)*Gamma(2.-2.*M))*(fxim/(1+fxim))*(1.+1./fxim-twoFone);
    }
    return result;// - singMellin(type, M);
  }
  struct Intpars {
    Interpolator *Int;
    double xi;
    string type;
  };
  double integrandC(double t, void *pars) {
    Intpars p = *(Intpars *) pars;
    Interpolator *MInt = p.Int;
    double xi = p.xi;
    string type = p.type;
    dcomplex M = MInt->f_reM0 + (MInt->f_slope-I)*log(t);
    dcomplex res = (-1.)*(I-MInt->f_slope) * exp((-1. - M)*log(xi)) * MInt->mellinC(type,M)/t;
    return imag(res)/M_PI;
  }
  double Interpolator::InverseMellinC(string type, double xi) {
    if (xi==0) return Csmallxi(type,0);
    if (xi==1./0) return 0;
    double res=0, prec = 1.e-5;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    Intpars pars = { this, xi, type };
    res = HELL::integration_gauss(integrandC,0.,1.,prec,NULL,&pars);
    gsl_set_error_handler (old_handler);
    return res;
  }














/*
int integrandC2CUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars)
{
  Intpars p = *(Intpars *) pars;
  Interpolator *Int = p.Int;
  double xi = p.xi;
  res[0] = Int->C2V1(x[0],x[1],xi);
  return 0;
}

double Interpolator::C2CUBA(double xi)
{
  double cres[1], error[1], prob[1];
  double epsrel=1.e-5, epsabs=1.e-100;
  int last = 4;
  int verbose = 0;
  int nregions, neval, fail;
  Intpars pars = { this, xi };
  Cuhre(2, 1, integrandC2CUBA, &pars, 1,
	epsrel, epsabs, verbose | last,
	0, 500000, 9, NULL, NULL,
	&nregions, &neval, &fail, cres, error, prob);
  if(fail == -1) {
    cout << "  ERROR: dimension out of range." << endl;
  } else if(fail>0) {
    cout << "  ERROR: precision not reached." << endl;
  }
  return cres[0];
}

double Interpolator::C2V1(double x1,double x2, double xi)
{
  double xim = 4*fxim;
  return (3*x1*((3 - 4*x1)*pow(x2,3) + 2*(-1 + x1)*pow(x2,2)*(3 + 4*x1*xi - xim) - 
	 2*pow(x1*xi + xim,2) + 
	 x2*(-(xim*(8 + xim)) + pow(x1,2)*xi*(-6 + xi + 2*xim) + 
	     2*x1*(2*xi + xim*(3 + xim)))))/
	     (8.*sqrt(1 - x1)*sqrt(1 - x2)*pow(x2 + x1*xi + xim,4)); 
}

int integrandC2masslessCUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars)
{
  Intpars p = *(Intpars *) pars;
  Interpolator *Int = p.Int;
  double xi = p.xi;
  res[0] = Int->C20(x[0],x[1],xi);
  return 0;
}

double Interpolator::C2masslessCUBA(double xi)
{
  double cres[1], error[1], prob[1];
  double epsrel=1.e-5, epsabs=1.e-100;
  int last = 4;
  int verbose = 0;
  int nregions, neval, fail;
  Intpars pars = { this, xi };
  Cuhre(2, 1, integrandC2masslessCUBA, &pars, 1,
	epsrel, epsabs, verbose | last,
	0, 500000, 9, NULL, NULL,
	&nregions, &neval, &fail, cres, error, prob);
  if(fail == -1) {
    cout << "  ERROR: dimension out of range." << endl;
  } else if(fail>0) {
    cout << "  ERROR: precision not reached." << endl;
  }
  return cres[0];
}

double Interpolator::C20(double x1,double x2, double xi)
{
  return 3./8*(x1*(pow(x2,2)*(-6. + x1*(6. - 4.*x2) + 3.*x2) + 
       x1*x2*(4. - 8.*x2 + x1*(-6. + 8.*x2))*xi + pow(x1,2)*(-2. + 1.*x2)*pow(xi,2)))
    /(sqrt(1 - x1)*sqrt(1 - x2)*pow(x2 + x1*xi,4));
}





int integrandCLmasslessCUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars)
{
  Intpars p = *(Intpars *) pars;
  Interpolator *Int = p.Int;
  double xi = p.xi;
  res[0] = Int->CL0(x[0],x[1],xi);
  return 0;
}

double Interpolator::CLmasslessCUBA(double xi)
{
  double cres[1], error[1], prob[1];
  double epsrel=1.e-5, epsabs=1.e-100;
  int last = 4;
  int verbose = 0;
  int nregions, neval, fail;
  Intpars pars = { this, xi };
  Cuhre(2, 1, integrandCLmasslessCUBA, &pars, 1,
	epsrel, epsabs, verbose | last,
	0, 500000, 9, NULL, NULL,
	&nregions, &neval, &fail, cres, error, prob);
  if(fail == -1) {
    cout << "  ERROR: dimension out of range." << endl;
  } else if(fail>0) {
    cout << "  ERROR: precision not reached." << endl;
  }
  return cres[0];
}

double Interpolator::CL0(double x1,double x2, double xi)
{
  return -3./8*((x1*pow(x2,2))/(sqrt(1 - x1)*sqrt(1 - x2)*pow(x1 + x2*xi,2)));
}
*/

};
