#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <cstdlib>

#include "Pqg.hh"
#include "math/borel.hh"
#include "commons.hh"
#include "gammaNLO.hh"
#include "h_coefficients.hh"
#include "CoefficientFunctions.hh"


using namespace std;



namespace HELL {


  unsigned int RCmodeOld = 0;
  void SetRCmodeCFold(unsigned int mode) {  // mode=0 uses gammadot, mode=1 uses as*beta0
    if(mode>1) {
      cout << "Error: invalid RC mode. Using default." << endl;
      mode = 0;
    }
    RCmodeOld = mode;
  }
  bool RCvarOld = 0;
  void SetRCvarCFold(bool var) {  // var=true uses "Aiy" anomalous dimension
    RCvarOld = var;
  }

  dcomplex hRC(dcomplex gamma, dcomplex gammadot, hCoeff hC, int nPade=8) {
    int Ncoeff = hC.hk().size();
    if(Ncoeff < 2*nPade) {
      cout << "Not enough coefficients" << endl;
      exit(345);
    }
    Ncoeff = 2*nPade;
    //
    dcomplex *coeff_ck   = new dcomplex[Ncoeff];
    //dcomplex *sqbr_gamma = new dcomplex[Ncoeff];
    dcomplex gammadot_over_gamma2 = gammadot/gamma/gamma;
    //sqbr_gamma[0] = 1.;
    dcomplex sqbr_gamma = 1.;
    coeff_ck[0] = hC.hk(0);
    for(int k=1; k<Ncoeff; k++) {
      //sqbr_gamma[k] = gamma * ( 1. + (k-1.)*gammadot_over_gamma2 ) * sqbr_gamma[k-1];
      //coeff_ck[k] = hC.hk(k) * sqbr_gamma[k];
      sqbr_gamma *= gamma * ( 1. + (k-1.)*gammadot_over_gamma2 );
      coeff_ck[k] = hC.hk(k) * sqbr_gamma;
      //cout << coeff_ck[k] << endl;
    }
    //double Lambda = 20.;
    //nPade = Ncoeff/2-1;
    SetBorelOrder(2);  // !!!!!! <<=== parameter here: works with 1 or 2 
    //dcomplex borelsum = BorelPadeSum(1., nPade, coeff_ck, Ncoeff);
    dcomplex borelsum = BorelPadeSum(1., nPade, nPade-1, coeff_ck, Ncoeff);
    //dcomplex borelsum = BorelPadeSum(1., nPade, 1, coeff_ck, Ncoeff);
    //dcomplex borelsum = BorelTruncatedSumCUBA (1., hqgNcoeff-1, Lambda, coeff_ck, Ncoeff);
    //dcomplex borelsum = BorelTruncatedSumGauss(1., hqgNcoeff-1, Lambda, coeff_ck, Ncoeff);
    //dcomplex borelsum = BorelTruncatedSumCUBA (1., 20, Lambda, coeff_ck, Ncoeff);
    //dcomplex borelsum = BorelTruncatedSumGauss(1., 13, Lambda, coeff_ck, Ncoeff); // Peraro like
    return borelsum;
  }


  dcomplex fdamp(dcomplex N) {
    return 1./(N+1.);
  }
  dcomplex fmom(dcomplex N) { // same form as used in preHELL
    //return 12.*N/pow(1.+N,2)/(2.+N);
    //return 8.*N/pow(1.+N,3);
    return 4.*N/pow(N+1.,2.);
  }


  bool a10woRC = false;
  bool fullySymm = false;
  bool shiftMmin = true;

  const double c0 = CA/M_PI*4*log(2.);
  const double k0 = CA/M_PI*28.*ZETA3;
  // from DL-LO kernel
  inline double c1LO (int nf) { return -15.00496429 - 0.04503163717*nf;}
  inline double k1LO (int nf) { return -507.744719  - 1.080759292*nf;}
  // from DL-NLO kernel
  //inline double c1NLO(int nf) { return -15.04043249 - 0.2076545137*nf; }
  //inline double k1NLO(int nf) { return -574.496771  - 0.3724353405*nf; }
  //inline double m1NLO(int nf) { return 0.3099881585 - 0.01878716112*nf; }
  inline double c1NLO(int nf) { return -11.696833425 - 0.4102968810*nf; }
  inline double k1NLO(int nf) { return -494.250393369 - 5.23585215538*nf; }
  inline double m1NLO(int nf) { return -0.1492429211 + 0.00904502552*nf; }
  //
  double a11()       { return CA/M_PI; }
  double a10(int nf) { return -(11.*CA + 2.*nf*(1.-2.*CF/CA))/12./M_PI + (a10woRC ? beta0(nf) : 0 ); }
  double a21(int nf) { return nf*(26*CF-23*CA)/36/M_PI/M_PI; }
  double a20(int nf) { return (19.6798969734649 + nf*(4./9.*ZETA2-68./81.) + nf*nf*13./2187.)/M_PI/M_PI; }
  double a32(int nf) { return -CA*( (18.*ZETA2-71.)*(CA-2*CF)*nf + CA*CA*(54*ZETA3+99.*ZETA2-395) ) /108./M_PI/M_PI/M_PI; }
  //
  dcomplex gamma0approx(dcomplex N, double nf) {
    double a1 = a11();
    double a0 = a10(nf) - (a10woRC ? beta0(nf) : 0 );
    return a1/N + a0 + 2*(a1+a0)*(1./(N+1.)-1.); // preserves momentum conservation, accurate at NLL, constant at large N
    //return gamma0pl(N+1., 0) + fdamp(N)*nf*(2*CF/CA-1)/6./M_PI;
  }
  dcomplex gamma0singNoDamp(dcomplex N, double nf) {
    double momcons = 0;
    double a1 = a11();
    double a0 = a10(nf) - (a10woRC ? beta0(nf) : 0 );
    return a1/N + a0*fdamp(N) - momcons*fmom(N)*(a1+a0*fdamp(1.)); // preserves momentum conservation, accurate at NLL, zero at large N
  }
  dcomplex gamma0sing(dcomplex N, double nf) {
    //return gamma0approx(N,nf);  // I'd discard this option, as a purely resummed result should vanish at large N, ideally...
    return gamma0singNoDamp(N,nf);
    //return gamma0singNoDamp(N,nf)/(N+1.);
    //return gamma0singNoDamp(N,nf) - 2.*gamma0singNoDamp(N+1.,nf) + gamma0singNoDamp(N+2.,nf);
  }
  dcomplex gamma1singLLp(dcomplex N, double nf) {
    double c0 = 4*log(2.);
    double k0 = 28*ZETA3;
    return beta0(nf)*CA/M_PI*(3*k0/32.-c0) * (1./N - fmom(N)); // add the correct as^2 contribution (from Bateman only - the DL part is zero)
    //return beta0(nf)*CA/M_PI*(-c0/4.) * (1./N - fmom(N)); // add the correct as^2 contribution (from Bateman only - the DL part is zero)
  }
  dcomplex gamma1singNLL(dcomplex N, double nf) {
    //double momcons = 0;
    double a1 = a21(nf);
    //double a0 = a20(nf);
    //return ( a1/N + a0*fdamp(N) - momcons*fmom(N)*(a1+a0*fdamp(1.)) ); // preserves momentum conservation, accurate at NNLL, zero at large N
    return a1/N - 2*a1/(N+1.); // preserves momentum conservation, accurate at NLL, zero at large N
  }
  dcomplex gamma1sing(dcomplex N, double nf) {
    return ( useLLp ? gamma1singLLp(N,nf) : gamma1singNLL(N,nf) );
  }

dcomplex dpsi(dcomplex Z, int M=1){
  int K1, K2;
  dcomplex SUB = 0. , SUBM;
  dcomplex ZZ = Z;
  if(fabs(imag(ZZ))<10.) { // if too close to the real axis...
  label1:
    SUBM = -1./ZZ;
    for(K1=1; K1<=M; K1++) {
      SUBM = - SUBM * (double)K1 / ZZ;
    }
    if(real(ZZ)<10.) { // ...use recurrence relation to push real(z) large enough
      SUB = SUB + SUBM;
      ZZ = ZZ + 1.;
      goto label1;
    }
  }
  // Expansion coefficients for the first derivative
  double A1 =  1.;
  double A2 =  1./2.;
  double A3 =  1./6.;
  double A4 = -1./30.;
  double A5 =  1./42.;
  double A6 = -1./30.;
  double A7 =  5./66.;
  // Expansion coefficients for the higher derivatives
  if(M>1) {
    for(K2=2;K2<=M;K2++){
      A1 = A1 * (1.*K2-1.);
      A2 = A2 *  1.*K2;
      A3 = A3 * (1.*K2+1.);
      A4 = A4 * (1.*K2+3.);
      A5 = A5 * (1.*K2+5.);
      A6 = A6 * (1.*K2+7.);
      A7 = A7 * (1.*K2+9.);
    }
  }
  dcomplex RZ = 1./ ZZ;
  dcomplex DZ = RZ * RZ;
  // SUB + asympt expansion (Abramowitz, Stengun, 6.4.11)
  return SUB + pow(-1.,M+1) * pow(RZ,M) * ( A1 + RZ * (A2 + RZ * (A3 + DZ * (A4 + DZ * (A5 + DZ * (A6 + A7 * DZ ))))) );
}
  dcomplex gamma_DLNLO_order_as3(dcomplex N, int nf) {
    double a1 = a11(), a0=a10(nf);
    double chi01 = a1;
    double chi02 = a0*a1;
    double chi11 = a21(nf);
    double rho = chi01*chi01*(5*ZETA3/2.-1./24.) + chi02*(53./18.-ZETA2) - chi11;
    //double rho = ( CA*CA*(-74./27. +11./12.*ZETA2 +5./2.*ZETA3) +nf*((4*CA+7*CF)/27. +(CA-2*CF)*ZETA2/6.) )/M_PI/M_PI;
    dcomplex gamma0  =  a1/N + a0 - 2*(a1+a0)*N/(N+1.);
    dcomplex gamma0p = -a1/N/N - 2*(a1+a0)/(N+1.)/(N+1.);
    double b0 = beta0(nf);
    dcomplex res = rho + chi11/(N+1.) + chi02/pow(N+1.,2) - (rho+chi11/2.+chi02/4.-b0*chi01)*fmom(N) + chi01*(dpsi(N+1.)-dpsi(1))*(gamma0+b0);
    if(a10woRC)   res -= b0 * chi01 * (dpsi(N+1.)-dpsi(1) + fmom(N));
    if(fullySymm) res -= b0 * chi01 * (dpsi(N+1.)-dpsi(1) + fmom(N));
    return -gamma0p * res;
  }
  dcomplex gamma_BatemanColl_minus_dcNLO_momcons_order_as3(dcomplex N, int nf) {
    double b0 = beta0(nf);
    double res = b0*k0/16.;
    return b0* res * (1./N - fmom(N));
  }
  dcomplex gammaNLL2(dcomplex N, int nf) {
    return gamma_DLNLO_order_as3(N,nf) + gamma_BatemanColl_minus_dcNLO_momcons_order_as3(N,nf);
  }
  /*
  dcomplex gammaNLL2simpl(dcomplex N, int nf) {
    dcomplex res = 0;
    double a1 = a11(), a0=a10(nf);
    double A = ( CA*CA*(-74./27. +11./12.*ZETA2 +5./2.*ZETA3) +nf*((4*CA+7*CF)/27. +(CA-2*CF)*ZETA2/6.) )/M_PI/M_PI;
    double chi11 = a21(nf);
    double chi02 = a1*a0;
    dcomplex gamma0p = -a1/N/N - 2*(a1+a0)/(N+1.)/(N+1.);
    res  = -gamma0p * ( A + chi11/(N+1.) + chi02/pow(N+1.,2) - (A+chi11/2.+chi02/4.)*fmom(N) );
    res += CA/M_PI* ( a1*a1*  ( 3*ZETA4/N -2*ZETA3/N/N -3./(N+1.) -1./(N+1.)/(N+1.) )
		      +a1*a0* (-2*ZETA3/N +2./(N+1.) +1./(N+1.)/(N+1.) )
		      +4*pow(a0 + a1,2)*              ( 3*ZETA4/(N+1.) -2*ZETA3/(N+1.)/(N+1.) + 1./pow(N+1.,5) )
		      -2*(a0*a0 + 4*a0*a1 + 3*a1*a1)* ( -2*ZETA3/(N+1.) +1./pow(N+1.,4) )
		      );
    // add RC
    double b0 = beta0(nf);
    res += b0*b0*k0/16.* (1./N - fmom(N));
    return res;
  }
  */
  dcomplex gamma2singLLp(dcomplex N, int nf, bool var) {
    dcomplex res = 0;
    double a1 = a11(), a0=a10(nf);
    double chi02 = a1*a0;
    dcomplex gamma0  =  a1/N + a0 - 2*(a1+a0)*N/(N+1.);
    dcomplex gamma0p = -a1/N/N - 2*(a1+a0)/(N+1.)/(N+1.);
    res  = -gamma0p * ( chi02/pow(N+1.,2) - chi02/4.*fmom(N) + CA/M_PI*(dpsi(N+1.)-dpsi(1))*gamma0 );
    // add RC
    double b0 = beta0(nf);
    int T = (var ? 1 : 2);
    double  c1 = c1NLO(nf);
    double  k1 = k1NLO(nf);
    double  m1 = m1NLO(nf);
    double dc1 = c1-c1LO(nf);
    double dk1 = k1-k1LO(nf);
    double dm1 = m1;
    if(shiftMmin) m1 += c0/2.;
    double lam2 = b0*(48*c0*k0-3*k0*k0-256*c0*c0)/256. + (16*c0*(dk1+6*k0*dm1)+k0*(16*dc1-3*dk1-15*k0*dm1))/512.;
    double lam1 = b0*(3*T*k1/32 -T*c1 + (b0+6*m1)*k0/16.);
    res += lam1/N + lam2/N/N - (lam1+lam2)*fmom(N);
    return res;
  }  
  dcomplex gamma2sing(dcomplex N, double nf, bool var) {
    return ( useLLp ? gamma2singLLp(N,nf,var) : gammaNLL2(N,nf) );
    //return ( useLLp ? gamma2singLLp(N,nf) : gammaNLL2simpl(N,nf) );
  }

  dcomplex DeltaFRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, process proc, int ordersub=0, int nPade=8, void *pars=NULL) {
    hCoeff hC(proc, 1);
    double h0 = hC.hk(0);
    double h1 = hC.hk(1);
    double h2 = hC.hk(2);
    double h3 = hC.hk(3);
    //
    double b0 = beta0(nf);
    dcomplex gamma0 = gamma0sing(N,nf);
    dcomplex gamma1 = gamma1sing(N,nf);
    //dcomplex gamma2 = (useLLp ? 0 : gamma2sing(N,nf,RCvarOld) );
    dcomplex gamma2 = gamma2sing(N,nf,RCvarOld);
    //
    //gamma0 = gamma0(N,nf) - gamma0(N+1.,nf);
    //gamma0 = gamma0(N,nf) - 2.*gamma0(N+1.,nf) + gamma0(N+2.,nf);
    //gamma0 = gamma0(N,nf) - 3.*gamma0(N+1.,nf) + 3.*gamma0(N+2.,nf) - gamma0(N+3.,nf);
    //*
    // Marco version
    dcomplex gammaLO_BatNLO    = Delta_gammaLO_BatNLO    + as*gamma0 + as*as*gamma1 + as*as*as*gamma2;
    dcomplex gammaLOdot_BatNLO = Delta_gammaLOdot_BatNLO +    gamma0 + 2.*as*gamma1 + 3.*as*as*gamma2;
    //
    //gammaLO_BatNLO /= N+1.; // hack! large-N damping
    //gammaLOdot_BatNLO /= N+1.; // hack! large-N damping
    if(RCmodeOld == 1) gammaLOdot_BatNLO = gammaLO_BatNLO/as;
    //
    gammaLOdot_BatNLO *= -b0*as*as;  // convert deriv wrt as to deriv wrt t
    //gammaLOdot_BatNLO = 0;
    dcomplex res = 0;
    if(proc == U) {
      double xi = *(double*)pars;
      dcomplex r = -gammaLOdot_BatNLO / gammaLO_BatNLO;
      res = evolU(xi, gammaLO_BatNLO, r);
      //res = Dxi_evolU(xi, gammaLO_BatNLO, r);
    } else {
      res = hRC(gammaLO_BatNLO, gammaLOdot_BatNLO, hC, nPade);
    }
    /*/
    // Richard version // not yet modified to account for the fact that Delta is now Delta2!!
    dcomplex FOgammaLO  = as*gamma0pl(N+1.,0);
    dcomplex gammaLO_BatNLO    = Delta_gammaLO_BatNLO    + FOgammaLO;
    dcomplex gammaLOdot_BatNLO = Delta_gammaLOdot_BatNLO + FOgammaLO/as;
    if(RCmodeOld == 1) gammaLOdot_BatNLO = gammaLO_BatNLO/as;
    gammaLOdot_BatNLO *= -b0*as*as;  // convert deriv wrt as to deriv wrt t
    FOgammaLO -= as*gamma0sing(N,0);
    dcomplex FOgammaLOdot  = -b0*as*FOgammaLO;
    //gammaLOdot_BatNLO = 0;
    //FOgammaLOdot = 0;
    dcomplex res = hRC(gammaLO_BatNLO, gammaLOdot_BatNLO, hC) - hRC(FOgammaLO, FOgammaLOdot, hC) + h0;
    //*/
    //
    //cout << res << endl;
    //if(res != 1.) exit(0);
    //
    //if(useLLp) gamma2 = gamma2sing(N,nf,RCvarOld); // recompute it since it was set to zero before
    double T = (RCmodeOld==1 ? 1 : 2);
    if(ordersub>0) res -= h0;
    if(ordersub>1) res -= h1*as*gamma0;
    if(ordersub>2) res -= as*as*(h2*gamma0*(gamma0-b0) + h1*gamma1);
    if(ordersub>3) res -= as*as*as*(h3*gamma0*(gamma0-b0)*(gamma0-2*b0) + h2*gamma1*(2.*gamma0-T*b0) + h1*gamma2);
    if(ordersub>4) cout << "Error, trying to subtract beyond the capability of HELL" << endl;
    //
    if(proc == logR) res *= gammaLOdot_BatNLO/gammaLO_BatNLO;
    //
    //cout << setw(15) << real(gammaLO_BatNLO) << setw(16) << real(res) << endl << endl;
    return res;
  }

  dcomplex Delta1Gamma(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex gamma1 = gamma1sing(N,nf);
    //dcomplex gamma2 = (useLLp ? 0 : gamma2sing(N,nf,RCvarOld) );
    dcomplex gamma2 = gamma2sing(N,nf,RCvarOld);
    return Delta_gamma + as*as*gamma1 + as*as*as*gamma2;
  }
  dcomplex Delta1GammaDot(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex gamma1 = gamma1sing(N,nf);
    //dcomplex gamma2 = (useLLp ? 0 : gamma2sing(N,nf,RCvarOld) );
    dcomplex gamma2 = gamma2sing(N,nf,RCvarOld);
    return Delta_gammadot + 2.*as*gamma1 + 3.*as*as*gamma2;
  }
  dcomplex Gamma0(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex gamma0 = gamma0sing(N,nf);
    return gamma0;
  }
  dcomplex Gamma0sq(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex gamma0 = gamma0sing(N,nf);
    return gamma0*gamma0;
  }
  dcomplex Gamma0cb(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex gamma0 = gamma0sing(N,nf);
    return gamma0*gamma0*gamma0;
  }
  dcomplex Gamma1(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex gamma1 = gamma1sing(N,nf);
    return gamma1;
  }
  dcomplex Gamma0Gamma1(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex gamma0 = gamma0sing(N,nf);
    dcomplex gamma1 = gamma1sing(N,nf);
    return gamma0*gamma1;
  }
  dcomplex Gamma2(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex gamma2 = gamma2sing(N,nf,RCvarOld);
    return gamma2;
  }

  dcomplex DeltaGammaResNLOqg(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = DeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, gamma_qg, 3, 8);
    return res * as*nf/3./M_PI;
  }
  dcomplex DeltaUqg(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = DeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Uqg, 1, 7); // should be 2, but then it's doubly counted in C2, so it's 1
    //dcomplex res = DeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Uqg, 0, 7);
    return res * as*nf/3./M_PI;
  }
  dcomplex DeltaUxi(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = DeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, U, 1, 7, pars);
    return res;
  }
  dcomplex DeltaLogR(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = DeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, logR, 3, 7);
    return res; // any factor?
  }
  dcomplex DeltaGammaPlusUqg(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    double b0 = beta0(nf);
    double c = (11*CA+2*nf*(1-2*CF/CA))/12.;
    dcomplex fixedorderpole = ( CA/N - c*fmom(N+1.) )/M_PI;
    dcomplex gammaLO_BatNLO    = Delta_gammaLO_BatNLO    + as*fixedorderpole;
    dcomplex gammaLOdot_BatNLO = Delta_gammaLOdot_BatNLO +    fixedorderpole;
    //gammaLOdot_BatNLO = gammaLO_BatNLO/as; //hack!!!
    gammaLOdot_BatNLO *= -b0*as*as;  // convert deriv wrt as to deriv wrt t
    dcomplex r = -gammaLOdot_BatNLO / gammaLO_BatNLO;
    dcomplex gamma0 = fixedorderpole;
    dcomplex gamma1 = nf*(26.*CF-23.*CA)/36./M_PI/M_PI/N;
    dcomplex gammaNLO = as*gamma0 + as*as*gamma1 + Delta_gammaNLO;
    hCoeff hC(Uqg,1);
    double h0 = hC.hk(0);
    double h1 = hC.hk(1);
    //dcomplex extraU  =  h0*as/(r + gammaLO_BatNLO);
    dcomplex extraU  =  h0*as/(r + gammaNLO);
    dcomplex extraU0 =  h0/(b0+gamma0);
    dcomplex extraU1 = -h0*gamma1/gamma0/(b0+gamma0);
    extraU = extraU0 = extraU1 = 0; // hack!
    dcomplex U  = nf/3./M_PI * ( as*DeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Uqg, 0, 7) + extraU );
    dcomplex U0 = nf/3./M_PI * ( extraU0 );
    dcomplex U1 = nf/3./M_PI * ( extraU1 + h1 );
    return gammaNLO*U - as*gamma0*U0 - as*as*(gamma1*U0 + gamma0*U1);
    return (gammaNLO+r)*U - as*(gamma0+b0)*U0 - as*as*(gamma1*(1.+b0/gamma0)*U0 + (gamma0+b0)*U1);
  }
  dcomplex oldDeltaCLRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = DeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, CL, 2, 8);
    return res * as*nf/3./M_PI;
  }
  dcomplex oldDeltaC2Res(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = DeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, C2, 2, 8);
    return res * as*nf/3./M_PI;
  }



};

